# TODO docs

from os.path import expanduser
from os.path import expandvars
from pathlib import Path

from dfmpy.utils.config import get_config
from dfmpy.utils.files.constants import suffixes


def normalize_for_phantom_files(file_paths):
    # TODO docs - explain this shit!  explain why it's needed when determining expected files when there only exists a markered file (and not the equivalent non-marked file).
    # TODO use this with EXTREME caution, since it's a hack it causes too many side-effects in other places -- causing misleading bugs.
    # TODO unit test
    # TODO optimize this? so it does not iterate over the list many times! (eg use of "in" keyword).
    file_paths = list(file_paths)
    marker = get_config().marker
    for fp in file_paths:
        file_name = str(fp)
        if marker in file_name:
            marker_index = file_name.find(marker)
            phantom_file_name = file_name[:marker_index]
            phantom_file_path = Path(phantom_file_name)
            if phantom_file_path not in file_paths:
                file_paths.append(phantom_file_path)
    file_paths.sort()
    return tuple(file_paths)


def normalize_repo_path(file_path):
    # TODO docs
    # TODO unit test
    file_name = str(file_path)
    for suffix in suffixes():
        file_with_suffix = file_name + suffix
        if Path(file_with_suffix).exists():
            return Path(file_with_suffix)
    return Path(file_path)


def normalize_file_names(files):
    # TODO docs
    # TODO unit test
    if not files:
        files = []
    files = [expanduser(expandvars(f)) for f in files]
    files = [Path(f).absolute() for f in files]
    return tuple(files)



# TODO docs


from collections import OrderedDict
from functools import lru_cache
from logging import getLogger
from os import R_OK
from os import access
from pathlib import Path

from dfmpy.utils.config import get_config
from dfmpy.utils.config import get_ignore_globs
from dfmpy.utils.files import path_is_directory
from dfmpy.utils.files import path_is_file
from dfmpy.utils.files import path_is_symlink
from dfmpy.utils.files.normalizer import normalize_for_phantom_files
from dfmpy.utils.files.normalizer import normalize_repo_path

# TODO docs
LOG = getLogger()

# TODO docs
# The cache should never get larger than 2, one for $HOME and one for the repo
# dir.  There should be no need for a cache of any other directory.
__FILE_TREE_CACHE_SIZE = 2


def __ignore(path):
    # TODO docs
    # TODO unit test
    for pattern in get_ignore_globs():
        if path.match(pattern):
            return True
    return False


def __no_permissions_to_stat(path):
    # TODO docs - the path must exist, but is not accessible.
    # TODO unit test
    # TODO is it possible to remove access() in favor or Path() methods?
    return path.exists() and not access(path, R_OK)


def __find_all_paths(dir_path):
    # TODO docs - returns all files and (directories with markers)
    # TODO unit test
    marker = get_config().marker
    if not dir_path.exists():
        return tuple()
    else:
        for path in dir_path.iterdir():
            if __ignore(path):
                LOG.debug('Ignoring: %s', path)

            elif __no_permissions_to_stat(path):
                LOG.error('No permissions to read: %s', path)

            elif path_is_directory(path):
                if marker in path.name:
                    LOG.debug('Found special directory: %s', path)
                    yield path
                else:
                    LOG.debug('Traversing down directory: %s', path)
                    yield from get_all_paths(path)

            elif path_is_file(path):
                LOG.debug('Found file: %s', path)
                yield path

            elif path_is_symlink(path):
                LOG.debug('Found symlink: %s', path)
                yield path

            else:
                LOG.error('Found unknown node type: %s', path)


@lru_cache(maxsize=__FILE_TREE_CACHE_SIZE)
def get_all_paths(dir_name):
    # TODO docs - returns the same as __find_all_paths()
    # TODO unit test
    paths = []
    paths.extend(__find_all_paths(Path(dir_name)))
    paths.sort()
    return tuple(paths)


@lru_cache(maxsize=__FILE_TREE_CACHE_SIZE)
def get_installed_paths(install_dir, repo_dir):
    # TODO docs
    # TODO unit test
    LOG.debug('Searching for installed paths under: %s', install_dir)
    installed_paths = get_all_paths(install_dir)
    installed_paths = [p for p in installed_paths if path_is_symlink(p)]
    installed_paths = [p for p in installed_paths
                       if str(p.resolve()).startswith(repo_dir)]
    mapping = OrderedDict()
    mapping.update([(p, Path(p.resolve())) for p in installed_paths])
    return mapping


@lru_cache(maxsize=__FILE_TREE_CACHE_SIZE)
def get_expected_paths(install_dir, repo_dir):
    # TODO docs
    # TODO unit test
    LOG.debug('Searching for expected paths under: %s', repo_dir)
    marker = get_config().marker
    repo_paths = get_all_paths(repo_dir)
    repo_paths = normalize_for_phantom_files(repo_paths)
    repo_paths = [p for p in repo_paths if not __ignore(p)]
    repo_paths = [p for p in repo_paths if marker not in p.name]
    mapping = OrderedDict()
    for rp in repo_paths:
        installed_file = str(rp).replace(repo_dir, install_dir)
        installed_path = Path(installed_file)
        nrp = normalize_repo_path(rp)
        if nrp.resolve().exists():
            mapping[installed_path] = nrp
    return mapping



# TODO docs

from datetime import datetime
from logging import getLogger
from pathlib import Path
from shutil import copyfile
from shutil import move

from pkg_resources import resource_filename
from xdgenvpy import XDGPedanticPackage

# TODO docs
LOG = getLogger()

# TODO consider renaming this to a "config" command, where it can be initialized or "edit" the configs.

# TODO docs - Using the pedantic object as we want to ensure the full directory paths up to the file exist!
__XDG = XDGPedanticPackage('dfmpy')

# TODO docs
_BACKUP_SUFFIX = datetime.now().strftime('.%Y%m%d_%H%M%S.dfmpy.backup')


def __install_file(resource_name, overwrite=False):
    # TODO docs
    # TODO unit test
    base_resource = resource_filename('dfmpy.resources', resource_name)
    LOG.debug('Found: %s', base_resource)

    installed_file = Path(__XDG.XDG_CONFIG_HOME).joinpath(resource_name)
    if installed_file.exists() and not overwrite:
        LOG.error('Not overwriting: %s', installed_file)

    elif installed_file.exists() and overwrite:
        # TODO create a backup of the existing file!!
        backup_file = str(installed_file) + _BACKUP_SUFFIX
        move(installed_file, backup_file)
        copyfile(base_resource, installed_file)
        LOG.warning('Overwriting: %s, backup: %s', installed_file, backup_file)

    else:
        copyfile(base_resource, installed_file)
        LOG.info('Installing: %s', installed_file)


def initialize(force=False, interactive=False):
    # TODO do not overwrite existing files
    # TODO docs
    # TODO unit test
    # TODO implement interactive
    if interactive:
        raise NotImplementedError('Interactive not yet implemented.')
    __install_file('config.ini', overwrite=force)
    __install_file('ignore.globs', overwrite=force)


def initialize_main(cli):
    # TODO docs
    # TODO unit test
    initialize(cli.force, cli.interactive)



# TODO package level docs

# TODO docs
__all__ = tuple([
    'mkdir_parents'
    'path_is_directory',
    'path_is_file',
    'path_is_symlink',
    'path_exists',
])

from logging import getLogger

from dfmpy.utils.files.normalizer import normalize_for_phantom_files
from dfmpy.utils.files.normalizer import normalize_repo_path

# TODO docs
LOG = getLogger()


def mkdir_parents(symlink_path):
    # TODO docs
    # TODO unit test
    if not symlink_path.parent.exists():
        LOG.warning('Making directory with parents: %s', symlink_path.parent)
        symlink_path.parent.mkdir(parents=True)


def path_is_directory(path):
    # TODO docs - explain why path.is_dir() alone is not safe.
    # TODO unit test
    return path_exists(path) and path.is_dir() and not path.is_symlink()


def path_is_file(path):
    # TODO docs - explain why path.is_file() alone is not safe.
    # TODO unit test
    return path_exists(path) and path.is_file()


def path_is_symlink(path):
    # TODO docs - explain why path.is_symlink() alone is not safe.
    # TODO unit test
    return path_exists(path) and path.is_symlink()


def path_exists(path):
    # TODO docs - explain why path.exists() alone is not safe.
    # TODO unit test
    try:
        # Stat the path, but do not follow symlinks.
        path.lstat()
        return True
    except FileNotFoundError:
        return False



#!/usr/bin/env python3

# TODO docs

from platform import python_version
from platform import python_version_tuple

# TODO unit test
__PYTHON_MAJOR_VERSION = int(python_version_tuple()[0])
if __PYTHON_MAJOR_VERSION < 3:
    raise SystemError('dfmpy requires Python 3.6+, not ' + python_version())

# TODO unit test
__PYTHON_MIN_VERSION = int(python_version_tuple()[1])
if __PYTHON_MIN_VERSION < 6:
    raise SystemError('dfmpy requires Python 3.6+, not ' + python_version())

import logging
from argparse import ArgumentParser
from logging import StreamHandler
from logging import getLogger
from sys import stdout

import pkg_resources

from dfmpy.commands.add import add_main
from dfmpy.commands.initialize import initialize_main
from dfmpy.commands.ls import ls_main
from dfmpy.commands.uninstall import uninstall_main
from dfmpy.commands.sync import sync_main
from dfmpy.utils.interactive import check_for_interactivity
from dfmpy.utils.logging import get_formatter

# TODO docs
LOG = getLogger()

# TODO docs
__LOG_LEVELS = tuple([
    logging.CRITICAL,
    logging.ERROR,
    logging.WARNING,
    logging.INFO,
    logging.DEBUG,
])


def get_version_str():
    # TODO docs
    # TODO unit test
    return '%(prog)s ' + pkg_resources.require('dfmpy')[0].version


def __setup_logger(verbosity_index=None):
    # TODO docs
    # TODO unit test
    if not verbosity_index:
        verbosity_index = 1
    verbosity_index = max(0, min(len(__LOG_LEVELS) - 1, verbosity_index))
    log_level = __LOG_LEVELS[verbosity_index]
    # formatter = Formatter('%(asctime)s %(levelname)s %(message)s')
    formatter = get_formatter('%(levelname)s %(message)s')
    handler = StreamHandler(stdout)
    handler.setFormatter(formatter)
    logger = getLogger()
    logger.setLevel(log_level)
    logger.addHandler(handler)


def __parse_arguments():
    # TODO docs
    # TODO unit test
    # TODO go through all "help" docs and make sure they are consistent.

    verbose_args = tuple(['-v', '--verbose'])
    verbose_kwargs = {
        'action': 'count',
        'default': 1,
        'dest': 'verbosity',
        'help': 'Set verbosity level, multiple flags increase the level.',
    }
    force_args = tuple(['-f', '--force'])
    force_kwargs = {
        'action': 'store_true',
        'default': False,
        'dest': 'force',
    }
    interactive_args = tuple(['-i', '--interactive'])
    interactive_kwargs = {
        'action': 'store_true',
        'default': False,
        'dest': 'interactive',
    }

    parser = ArgumentParser(prog='dfmpy')
    main_verbose_kwargs = verbose_kwargs.copy()
    del main_verbose_kwargs['dest']
    parser.add_argument(*verbose_args,
                        **main_verbose_kwargs,
                        dest='main_verbosity')
    parser.add_argument('-V',
                        '--version',
                        action='version',
                        version=get_version_str())
    subparsers = parser.add_subparsers(help='sub-command help')

    add_parser = subparsers.add_parser('add', help='add file help')
    add_parser.set_defaults(func=add_main)
    add_parser.add_argument(*verbose_args, **verbose_kwargs)
    add_parser.add_argument(*force_args,
                            **force_kwargs,
                            help='Force overwriting of files.')
    add_parser.add_argument(*interactive_args,
                            **interactive_kwargs,
                            help='Interactively add files to the dotfiles repo,'
                                 ' and then sync it.')
    add_parser.add_argument('-F',
                            '--files',
                            dest='files',
                            help='Add one or more files to the dotfiles repo, '
                                 'then sync it.',
                            nargs='+')

    init_parser = subparsers.add_parser('init', help='initialize help')
    init_parser.set_defaults(func=initialize_main)
    init_parser.add_argument(*verbose_args, **verbose_kwargs)
    init_parser.add_argument(*force_args,
                             **force_kwargs,
                             help='Force overwriting of files.')
    init_parser.add_argument(*interactive_args,
                             **interactive_kwargs,
                             help='Interactively overwrite files if there are'
                                  ' conflicts.')

    ls_parser = subparsers.add_parser('list', help='list help')
    ls_parser.set_defaults(func=ls_main)
    ls_parser.add_argument(*verbose_args, **verbose_kwargs)

    sync_parser = subparsers.add_parser('sync', help='sync help')
    sync_parser.set_defaults(func=sync_main)
    sync_parser.add_argument(*verbose_args, **verbose_kwargs)
    sync_parser.add_argument(*force_args,
                             **force_kwargs,
                             help='Overwrite files if there are conflicts.')
    sync_parser.add_argument(*interactive_args,
                             **interactive_kwargs,
                             help='Interactively overwrite files if there are'
                                  ' conflicts.')

    uninstall_parser = subparsers.add_parser('uninstall', help='uninstall help')
    uninstall_parser.set_defaults(func=uninstall_main)
    uninstall_parser.add_argument(*verbose_args, **verbose_kwargs)
    uninstall_parser.add_argument(*force_args,
                                  **force_kwargs,
                                  help='Force removal of all files.')
    uninstall_parser.add_argument(*interactive_args,
                                  **interactive_kwargs,
                                  help='Interactively remove of files.')

    args = parser.parse_args()
    if not hasattr(args, 'func'):
        parser.print_help()
        parser.exit()

    # Since a argument cannot exist on both the main parser as well as a
    # sub-parser we need to normalize here.  They technically can both exist,
    # but the sub-parser overwrites any value calculated from parsing the main
    # arguments.
    if 'main_verbosity' in args:
        args.verbosity += args.main_verbosity - 1

    return args


def main():
    # TODO docs
    # TODO unit test?
    args = __parse_arguments()
    __setup_logger(args.verbosity)
    try:
        LOG.debug('Arguments: %s', args)

        if 'interactive' in args \
                and args.interactive \
                and not check_for_interactivity():
            raise RuntimeError(
                'Cannot run interactively while not attached to a TTY.')

        args.func(args)

    except KeyboardInterrupt:
        print('')
        LOG.warning('Leaving dfmpy prematurely.')

    # except Exception as e:
    #     LOG.critical(e)
    #     LOG.exception(e)

    except BaseException as e:
        # LOG.critical(e)
        LOG.exception(e)


if __name__ == '__main__':
    main()



# TODO docs

from logging import Formatter
from sys import stdout

from termcolor import colored


class NonColoredFormatter(Formatter):
    # TODO docs

    def format(self, record):
        # TODO docs
        # TODO unit test
        record.levelname = '[' + record.levelname[0] + ']'
        formatted_msg = super(NonColoredFormatter, self).format(record)
        return formatted_msg


class ColoredFormatter(NonColoredFormatter):
    # TODO docs

    # TODO docs - sub-dict args should match colored() args
    _COLORS = {
        'CRITICAL': {'color': 'white',
                     'on_color': 'on_red',
                     'attrs': tuple(['bold'])},
        'FATAL': {'color': 'white',
                  'on_color': 'on_red',
                  'attrs': tuple(['bold'])},
        'ERROR': {'color': 'red'},
        'WARNING': {'color': 'yellow'},
        'INFO': {'color': 'cyan'},
        'DEBUG': {},
    }

    def format(self, record):
        # TODO docs
        # TODO unit test
        # TODO move from termcolor to coloraama?
        # https://pypi.org/project/termcolor/
        # https://pypi.org/project/colorama/
        colors = ColoredFormatter._COLORS[record.levelname]
        formatted_msg = super(ColoredFormatter, self).format(record)
        colored_msg = colored(formatted_msg, **colors)
        return colored_msg


def get_formatter(*args, **kwargs):
    # TODO docs
    # TODO unit test
    if stdout.isatty():
        return ColoredFormatter(*args, **kwargs)
    else:
        return NonColoredFormatter(*args, **kwargs)



# TODO docs

from logging import getLogger

from dfmpy.utils.config import get_config
from dfmpy.utils.files.deleter import unlink_path
from dfmpy.utils.files.finder import get_installed_paths

# TODO docs
LOG = getLogger()


def uninstall(force=False, interactive=False):
    # TODO docs
    # TODO unit test

    install_dir = get_config().install_dir
    repository_dir = get_config().repository_dir

    installed_paths = get_installed_paths(install_dir, repository_dir)

    if not force:
        LOG.error("Must use '-f' to force removal of dotfiles.")

    for file in installed_paths:
        unlink_path(file,
                    force,
                    interactive,
                    tuple([install_dir, repository_dir]),
                    )


def uninstall_main(cli):
    # TODO docs
    # TODO unit test
    uninstall(cli.force, cli.interactive)



# TODO docs

from configparser import ConfigParser
from configparser import DEFAULTSECT
from configparser import NoSectionError
from functools import lru_cache
from logging import getLogger
from os.path import expanduser
from os.path import expandvars
from pathlib import Path
from platform import system
from socket import gethostname

from xdgenvpy import XDGPackage

# TODO add defaults in here so that the config files are not required.

# TODO docs
LOG = getLogger()

# TODO docs
__XDG_ENV = XDGPackage('dfmpy')

# TODO docs
__CONFIG_DEFAULTS = {
    'hostname': gethostname(),
    'system': system(),
}


class DfmpyConfig(object):
    # TODO docs

    def __init__(self, config):
        # TODO docs
        # TODO unit test
        self._config = config
        LOG.debug('Config: %s', dict(self._config))

    def __getattr__(self, attribute):
        # TODO docs
        # TODO unit test
        # TODO check CLI args for an override before checking the file
        value = self._config.get(attribute, None)

        # Normalize the value.
        if value:
            value = expanduser(expandvars(value.strip()))
        return value


def __remove_comments(s):
    # TODO docs
    # TODO unit test
    try:
        return s[:s.index('#')]
    except ValueError:
        return s


@lru_cache(maxsize=1)
def get_config():
    # TODO docs
    # TODO unit test
    # TODO supply a default set of args that was written to a default config file.
    config = ConfigParser(defaults=__CONFIG_DEFAULTS.copy())

    config_file = Path(__XDG_ENV.XDG_CONFIG_HOME).joinpath('config.ini')
    if config_file.exists():
        LOG.debug('Reading config file: %s', config_file)
        config.read(config_file)

    valid_sections = ('dfmpy',)
    section = DEFAULTSECT
    for vs in valid_sections:
        if config.has_section(vs):
            section = vs
    if not section:
        raise NoSectionError(f'Config {config_file}'
                             f' must include a valid section: {valid_sections}')
    LOG.debug('Using section %s from %s', section, config_file)
    try:
        return DfmpyConfig(config[section])
    except NoSectionError as e:
        e.message = f'{e.message} in {config_file}'
        raise e


def get_globs(filepath):
    # TODO docs
    # TODO unit test
    globs = []
    if Path(filepath).exists():
        with open(filepath, 'r') as f:
            globs = f.readlines()
        globs = [__remove_comments(line) for line in globs]
        globs = [line.strip() for line in globs]
        globs = [line for line in globs if line]
    return tuple(globs)


@lru_cache(maxsize=1)
def get_ignore_globs():
    # TODO docs
    # TODO unit test
    ignore_file = Path(__XDG_ENV.XDG_CONFIG_HOME).joinpath('ignore.globs')
    globs = get_globs(ignore_file)
    return tuple(globs)



# TODO docs

from logging import getLogger
from sys import stdin
from sys import stdout

# TODO docs
LOG = getLogger()


def check_for_interactivity():
    # TODO docs - can't be interactive if there is no terminal to read from or write to!
    # TODO unit test
    return stdout.isatty() and stdin.isatty()


def __prompt_user(message):
    # TODO docs - does not actually check "no" responses, just for a "yes."
    # TODO unit test
    # TODO make the y/N options as method arguments
    # TODO make a "strict" method param that loops this method if a y/N option is not selected.
    response = input(f'{message} [y/N]: ').strip().lower()
    response = (response and response[0] == 'y')
    if response:
        LOG.debug('User accepted the prompt: %s', message)
    else:
        LOG.debug('User DID NOT accepted the prompt: %s', message)
    return response


def ask_to(message, force=False, interactive=True):
    # TODO docs
    # TODO unit test

    if force:
        # If "force" is true, then allow the operation.
        return True

    elif not interactive:
        # If "force" is not true, and we do not want to ask the user, then for
        # safety do not allow the operation.
        return False

    else:
        # If "force" is not true, and we want to prompt the user, then ask them
        # what to do.
        return __prompt_user(message)



# TODO docs

from logging import getLogger

from dfmpy.utils.config import get_config
from dfmpy.utils.files.finder import get_expected_paths
from dfmpy.utils.files.finder import get_installed_paths

# TODO docs
LOG = getLogger()


def __unique_key_for_value(items, find_value):
    # TODO docs
    # TODO unit test
    keys = tuple([key for key, value in items if value == find_value])
    if not keys:
        return None
    elif len(keys) != 1:
        raise LookupError(f'Multiple keys for value: {find_value} => {keys}')
    return keys[0]


def __get_broken_paths(expected_paths, installed_paths):
    # TODO docs - The installed file points to a non-existent file.
    # TODO unit test
    LOG.debug('Finding installed files that point to non-existent files.')
    for f in installed_paths:
        if not f.resolve().exists():
            yield f, f.resolve()


def __get_stale_paths(expected_paths, installed_paths):
    # TODO docs - The installed file resolves to the wrong repo file.
    # TODO unit test
    LOG.debug('Finding installed files that resolve to the wrong file.')
    for f in installed_paths:
        expected = __unique_key_for_value(expected_paths.items(), f.resolve())
        if f != expected:
            yield f, expected


def __get_not_installed_paths(expected_paths, installed_paths):
    # TODO docs - The expected file is not pointed to by an installed file.
    # TODO unit test
    LOG.debug('Finding expected files that are not installed.')
    installed_files_keys = installed_paths.keys()
    for f in expected_paths.keys():
        if f not in installed_files_keys:
            yield f, expected_paths[f]


def __get_proper_paths(expected_paths, installed_paths):
    # TODO docs - The expected file is properly installed.
    # TODO unit test
    # TODO this needs testing once "sync" is implemented!
    LOG.debug('Finding installed properly installed files.')
    installed_files_keys = installed_paths.keys()
    for f in expected_paths:
        if f in installed_files_keys:
            installed = installed_paths[f]
            if f.resolve() == installed.resolve():
                yield f, expected_paths[f]


def __print(logger, files, message, comparator_symbol):
    # TODO docs
    # TODO unit test
    if files:
        for idx, (left, right) in enumerate(files):
            if idx == 0:
                # Since "files" could be a generator, only log the message if we
                # can actually iterate over the items.  Thus only print for the
                # first element.
                logger(message)
            logger('%s %s %s', left, comparator_symbol, right)


def ls():
    # TODO docs
    # TODO unit test
    # TODO add a "--tree" option to the this command

    install_dir = get_config().install_dir
    repository_dir = get_config().repository_dir

    expected_paths = get_expected_paths(install_dir, repository_dir)
    installed_paths = get_installed_paths(install_dir, repository_dir)

    broken = __get_broken_paths(expected_paths, installed_paths)
    __print(LOG.critical, broken, 'Found broken files:', '?=')

    stale = __get_stale_paths(expected_paths, installed_paths)
    __print(LOG.error, stale, 'Found stale files:', '~=')

    not_installed = __get_not_installed_paths(expected_paths, installed_paths)
    __print(LOG.warning, not_installed, 'Found files not installed:', '!=')

    proper = __get_proper_paths(expected_paths, installed_paths)
    __print(LOG.info, proper, 'Found installed files:', '==')


def ls_main(cli):
    # TODO docs
    # TODO unit test
    ls()



# TODO docs

from dfmpy.utils.config import get_config


def suffixes():
    # TODO docs
    # TODO unit test
    # Order of precedence:
    #   file.txt##hostname
    #   file.txt##hostname.system
    #   file.txt##systemname
    #   file.txt
    marker = get_config().marker
    delimiter = get_config().delimiter
    hostname = get_config().hostname
    system_type = get_config().system
    return tuple([
        marker + hostname,
        marker + delimiter.join([hostname, system_type]),
        marker + system_type,
    ])



# TODO package level docs

from dfmpy.commands.add import add
from dfmpy.commands.initialize import initialize
from dfmpy.commands.ls import ls
from dfmpy.commands.uninstall import uninstall
from dfmpy.commands.sync import sync

# TODO docs
__all__ = tuple([
    'add'
    'initialize',
    'ls',
    'uninstall',
    'sync',
])



# TODO docs

from logging import getLogger

from dfmpy.utils.files import mkdir_parents
from dfmpy.utils.files import path_is_file
from dfmpy.utils.files import path_is_symlink
from dfmpy.utils.files.deleter import unlink_path
from dfmpy.utils.interactive import ask_to

# TODO docs
LOG = getLogger()


def link_path(symlink_path, target_path, force, interactive):
    # TODO docs
    # TODO unit test
    if ask_to(f'Overwrite to link file: {symlink_path}', force, interactive):
        mkdir_parents(symlink_path)
        symlink_path.symlink_to(target_path)
        LOG.warning('Symlinked: %s -> %s', symlink_path, target_path)
    else:
        LOG.info('Simulated symlink: %s -> %s', symlink_path, target_path)


def sync_expected_paths(expected_paths, force, interactive):
    # TODO docs
    # TODO unit test
    LOG.info('Syncing files.')
    force_it_kwargs = {'force': True, 'interactive': False}
    for symlink_path, target_path in expected_paths.items():

        if path_is_symlink(symlink_path):
            if symlink_path.resolve() == target_path:
                # The symlink exists and points to the correct target.
                LOG.debug('No need to sync: %s -> %s',
                          symlink_path,
                          target_path)
            else:
                # If the symlink exists, but it points to the wrong target!
                sync_prompt = f'Sync broken link: {symlink_path}'
                if ask_to(sync_prompt, force, interactive):
                    unlink_path(symlink_path, **force_it_kwargs)
                    link_path(symlink_path, target_path, **force_it_kwargs)

        elif path_is_file(symlink_path):
            sync_prompt = f'Replace existing file: {symlink_path}'
            if ask_to(sync_prompt, force, interactive):
                unlink_path(symlink_path, **force_it_kwargs)
                link_path(symlink_path, target_path, **force_it_kwargs)

        else:
            # The file/link does not exist, so create it.
            link_path(symlink_path, target_path, force, interactive)



# TODO docs

from logging import getLogger
from pathlib import Path

from dfmpy.utils.files import path_is_directory
from dfmpy.utils.files import path_is_file
from dfmpy.utils.files import path_is_symlink
from dfmpy.utils.interactive import ask_to

# TODO docs
LOG = getLogger()


def __delete_empty_directories(file_path, root_names):
    # TODO docs
    # TODO unit test
    root_paths = [Path(r) for r in root_names]
    for p in file_path.parents:
        if p in root_paths:
            # Exit this loop to finish quickly since we hit a root.
            break
        elif not list(p.iterdir()):
            # Directory is empty, delete it.
            p.rmdir()
            LOG.warning('Unlinked empty directory: %s', p)
        else:
            # Exit early since we hit a non-empty directory.
            LOG.debug('Parent directory not empty, not unlinking: %s', p)
            break


def unlink_path(path, force, interactive, roots=None):
    # TODO docs
    # TODO unit test
    if not roots:
        roots = ('',)

    if ask_to(f'Unlink path: {path}', force, interactive):
        if path_is_directory(path):
            path.rmdir()
            LOG.warning('Removed directory: %s', path)
        elif path_is_file(path):
            path.unlink()
            LOG.warning('Removed file: %s', path)
        elif path_is_symlink(path):
            path.unlink()
            LOG.warning('Removed symlink: %s', path)
        else:
            path.unlink()
            LOG.error('Removed unknown file type: %s', path)
        __delete_empty_directories(path, roots)

    else:
        LOG.info('Simulated unlink: %s -> %s', str(path), path.resolve())


def remove_broken_files(installed_paths, force, interactive):
    # TODO docs
    # TODO unit test
    LOG.info('Checking for broken files.')
    for p in installed_paths:
        if not p.resolve().exists():
            unlink_path(p, force, interactive)



# /usr/bin/env python3

from functools import lru_cache
from pathlib import Path
from shutil import rmtree

from setuptools import Command
from setuptools import find_packages
from setuptools import setup
from setuptools.command.install import install

PROJECT_NAME = 'dfmpy'
PROJECT_VERSION = '0.0.4'


@lru_cache(maxsize=2)
def get_repo_file_name(filename):
    """
    Returns the filename that is relative to the project directory as an
    absolute file path.

    :param Path|str filename: The relative file path.

    :rtype: str
    :return: The absolute path to the specified file.
    """
    project_directory = Path(__file__).resolve().parent
    filename = project_directory.joinpath(filename)
    return filename


@lru_cache(maxsize=2)
def read_repo_file(filename):
    """
    Reads the specified file and returns the full contents as a single string.

    :param Path|str filename: The file to read.

    :rtype: str
    :return: The full contents of the specified file.
    """
    filename = get_repo_file_name(filename)
    with open(filename) as f:
        return f.read()


def load_requirements(filename):
    """
    Loads the :code:`requirements.tx` dependency file and returns the
    dependencies as a sequence.

    The file can contain comments that follow the pound symbol (eg. '#').
    Comments will be removed, each line is stripped of leading and trailing
    whitespace, and empty lines are deleted.  Other than these basic
    transformations, the requirements are left intact.

    :param Path|str filename: The requirements file to read.

    :rtype: tuple
    :return: A sequence of requirements.
    """
    reqs = read_repo_file(filename).splitlines()
    reqs = [str(x).strip() for x in reqs]
    reqs = [x[:x.find('#')] for x in reqs if '#' in x]
    reqs = [x for x in reqs if len(x)]
    return tuple(reqs)


def get_dfmpy_packages():
    """
    Finds all packages within this project and only returns the production ready
    ones.  Meaning, test packages will not be included.

    :rtype tuple
    :return: A sequence of package names that will be built into the file
            distribution.
    """
    packages = find_packages()
    packages = [p for p in packages if not p.endswith('_test')]
    return tuple(packages)


class CleanCommand(Command):
    """
    A custom clean command that removes any intermediate build directories.
    """

    description = 'Custom clean command that forcefully removes build, dist,' \
                  ' and other similar directories.'
    user_options = []

    def __init__(self, *args, **kwargs):
        """Initialized the custom clean command with a list of directories."""
        super(CleanCommand, self).__init__(*args, **kwargs)
        project_path = Path(__file__).resolve().parent
        self._clean_paths = {'.eggs',
                             'build',
                             PROJECT_NAME + '.egg-info',
                             'dist',
                             }
        self._clean_paths = {project_path.joinpath(p)
                             for p in self._clean_paths}
        self._clean_paths = {d for d in self._clean_paths if d.exists()}

    def initialize_options(self):
        """Unused, but required when implementing :class:`Command`."""
        pass

    def finalize_options(self):
        """Unused, but required when implementing :class:`Command`."""
        pass

    def run(self):
        """Performs the actual removal of the intermediate build directories."""
        for d in self._clean_paths:
            print(f'Removing {d}')
            rmtree(d)


class InstallCommand(install):
    """
    Custom command to install dependencies cause setuptools sucks balls and does
    not honor the "install_requires" parameter.  Apparently the default
    functionality is to run in backward-compatibile mode, which does not install
    the project's dependencies  :-(

    @see https://github.com/pypa/setuptools/issues/456#issuecomment-202922033
    """
    def run(self):
        """
        Performs the installation of dependencies, based on the
        :code:`requirements.txt` and :code:`requirements-test.txt` files.
        """
        # Need to run the original "install" first.
        super().run()

        # Not sure why, but moving these up to the top-level, and use the
        # "from PACKAGE import NAME" format, did not work.
        import subprocess
        import sys
        files = ('requirements.txt', 'requirements-test.txt')
        files = [get_repo_file_name(f) for f in files]
        for req_file in files:
            print(f'Installing dependencies from: {req_file}')
            process = subprocess.Popen(['pip3', 'install', '-r', req_file],
                                       stdout=subprocess.PIPE)
            for c in iter(lambda: process.stdout.read(1), b''):
                sys.stdout.write(c.decode('UTF-8'))


setup(name=PROJECT_NAME,
      version=PROJECT_VERSION,
      author='Mike Durso',
      author_email='rbprogrammer@gmail.com',
      classifiers=[
          'Environment :: Console',
          'Intended Audience :: Developers',
          'Intended Audience :: System Administrators',
          'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
          'Natural Language :: English',
          'Operating System :: MacOS',
          'Operating System :: POSIX',
          'Operating System :: POSIX :: Linux',
          'Operating System :: Unix',
          'Operating System :: Microsoft :: Windows',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3',
          'Topic :: Utilities',
      ],
      cmdclass={
          'clean': CleanCommand,
          'install': InstallCommand,
      },
      description='Another dotfiles manager.',
      entry_points={
          'console_scripts': [
              'dfmpy = dfmpy.__main__:main',
          ]
      },
      install_requires=load_requirements('requirements.txt'),
      long_description=read_repo_file('README.md'),
      long_description_content_type='text/markdown',
      packages=get_dfmpy_packages(),
      package_data={'dfmpy.resources': ['*']},
      test_suite='dfmpy_test',
      tests_require=load_requirements('requirements-test.txt'),
      url='https://gitlab.com/rbprogrammer/dfmpy',
      )



# TODO docs

from logging import getLogger
from pathlib import Path

from dfmpy.utils.files.normalizer import normalize_file_names
from dfmpy.utils.files.syncer import sync_expected_paths
from dfmpy.utils.config import get_config
from dfmpy.utils.files import mkdir_parents
from dfmpy.utils.files import path_exists

LOG = getLogger()


def __filter_non_destination_paths(paths):
    # TODO docs
    # TODO unit test
    install_dir = get_config().install_dir
    good_paths = []
    for p in paths:
        if not path_exists(p):
            LOG.error('Cannot add non-existent path: %s', p)
        elif not str(p).startswith(install_dir):
            LOG.error('Cannot add path not under destination: %s', p)
        else:
            good_paths.append(p)
    return tuple(good_paths)


def __determine_expected_paths(paths):
    # TODO docs
    # TODO unit test
    repository_dir = get_config().repository_dir
    install_dir = get_config().install_dir

    marker = get_config().marker
    delimiter = get_config().delimiter
    hostname = get_config().hostname
    system_type = get_config().system

    expected_paths = {}
    for p in paths:
        target_path = str(p).replace(install_dir, repository_dir)
        target_path += marker
        target_path += delimiter.join(hostname, system_type)
        expected_paths[p] = Path(target_path)
    return expected_paths


def __move_expected_paths(expected_paths, force):
    # TODO docs
    # TODO unit test
    for old, new in expected_paths.items():
        if not force:
            LOG.info('Simulated renaming %s -> %s', old, new)
        else:
            mkdir_parents(new)
            LOG.info('Renaming %s -> %s', old, new)
            Path(old).rename(new)


def add(files=None, force=False, interactive=False):
    # TODO do not overwrite existing files
    # TODO docs
    # TODO unit test
    # TODO implement interactive
    # TODO implement force
    paths = normalize_file_names(files)
    paths = __filter_non_destination_paths(paths)
    expected_paths = __determine_expected_paths(paths)
    __move_expected_paths(expected_paths, force)
    # TODO make this the next method public, or move to the "files" utility?
    sync_expected_paths(expected_paths, force, interactive)
    # if interactive:
    #     raise NotImplementedError('Interactive not yet implemented.')
    # __install_file('config.ini', overwrite=force)
    # __install_file('ignore.globs', overwrite=force)


def add_main(cli):
    # TODO docs
    # TODO unit test
    add(cli.files,
        cli.force,
        cli.interactive)



# TODO docs

from logging import getLogger

# TODO remove dependency on other commands
from dfmpy.utils.config import get_config
from dfmpy.utils.files.deleter import remove_broken_files
from dfmpy.utils.files.finder import get_expected_paths
from dfmpy.utils.files.finder import get_installed_paths
from dfmpy.utils.files.syncer import sync_expected_paths

# TODO docs
LOG = getLogger()


def sync(force=False, interactive=False):
    # TODO docs
    # TODO unit test

    install_dir = get_config().install_dir
    repository_dir = get_config().repository_dir

    expected_paths = get_expected_paths(install_dir, repository_dir)
    installed_paths = get_installed_paths(install_dir, repository_dir)

    if not force:
        LOG.error("Must use '-f' to force overwriting of symlinked dotfiles.")

    sync_expected_paths(expected_paths, force, interactive)
    remove_broken_files(installed_paths, force, interactive)


def sync_main(cli):
    # TODO docs
    # TODO unit test
    sync(cli.force, cli.interactive)
